/* powmod-k8-fpu.S -- (C) Geoffrey Reynolds, February 2007.

   uint64_t powmod64_k8_fpu(uint64_t b, uint64_t n, uint64_t p);

     Returns b^n (mod p), where 0 <= a < p < 2^62.

     Assumes FPU is set to double extended precision and round to zero.
     Assumes %st(0) contains 1.0/p computed with above settings.
     Uses red zone.


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include "config.h"

#if NEED_UNDERSCORE
#define powmod64_k8_fpu _powmod64_k8_fpu
#endif

#ifdef _WIN64
/* Shadow space */
#define VAR1 24(%rsp)
#define VAR2 32(%rsp)
#else
/* Red zone */
#define VAR1 -8(%rsp)
#define VAR2 -16(%rsp)
#endif

	.text
	.p2align 4
	.global powmod64_k8_fpu

powmod64_k8_fpu:
#ifdef _WIN64
	push	%rdi
	push	%rsi
	mov	%rcx, %rdi
	mov	%rdx, %rsi
	mov	%r8, %rdx
#endif
	mov	$1, %eax
	cmp	%rax, %rsi
	jbe	skip
	mov	%rdi, VAR1
	test	%eax, %esi
	fildll  VAR1
	cmovnz	%rdi, %rax
	jmp	sqr
skip:
	cmovz	%rdi, %rax
#ifdef _WIN64
	pop	%rsi
	pop	%rdi
#endif
	ret

	/* %rax = a, %rdi = b, %rsi = n, %rdx = p, %st(0) = 1.0/p */
	.p2align 4
loop:
	mov	%rdi, VAR1
	mov	%rax, VAR2
	test	$1, %sil
	fildll  VAR1
	jz	sqr
mul:
	fildll  VAR2
	fmul    %st(1), %st(0)
	fmul    %st(2), %st(0)
	fistpll VAR2
	imul    %rdi, %rax
	mov     VAR2, %rcx
	imul    %rdx, %rcx
	sub     %rcx, %rax
	mov     %rax, %rcx
	sub     %rdx, %rcx
	jl      sqr
	mov     %rcx, %rax
sqr:
	fmul    %st(0), %st(0)
	fmul    %st(1), %st(0)
	fistpll VAR2
	imul    %rdi, %rdi
	shr	%rsi
	mov     VAR2, %rcx
	imul    %rdx, %rcx
	sub     %rcx, %rdi
	mov     %rdi, %rcx
	sub     %rdx, %rcx
	jl	test
	mov     %rcx, %rdi
test:
	test	%rsi, %rsi
	jnz	loop

#ifdef _WIN64
	pop	%rsi
	pop	%rdi
#endif
	ret
