/* powmod-k8.S -- (C) Geoffrey Reynolds, February 2007.

   uint64_t powmod64_k8(uint64_t b, uint64_t n, uint64_t p, double invp);

     Returns b^n (mod p), where 0 <= b < p < 2^51.

     Assumes invp = 1.0/p (computed in round-to-zero mode).
     Assumes current SSE rounding mode is round-to-zero.


   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
*/

#include "config.h"

#if NEED_UNDERSCORE
#define powmod64_k8 _powmod64_k8
#endif

	.text
	.p2align 4
	.global powmod64_k8

powmod64_k8:
#ifdef _WIN64
	push	%rdi
	push	%rsi
	mov	%rcx, %rdi
	mov	%rdx, %rsi
	mov	%r8, %rdx
	movdqa	%xmm3, %xmm0
#endif
	mov	$1, %eax
	cmp	%rax, %rsi
	jbe	skip
	cvtsi2sd %rdi, %xmm1
	test	%eax, %esi
	cmovnz	%rdi, %rax
	jmp	sqr
skip:
	cmovz	%rdi, %rax
#ifdef _WIN64
	pop	%rsi
	pop	%rdi
#endif
	ret

	/* %rax = a, %rdi = b, %rsi = n, %rdx = p, %xmm0 = 1.0/p */
	.p2align 4
loop:
	test	$1, %sil
	cvtsi2sd %rdi, %xmm1
	jz	sqr
mul:
	cvtsi2sd %rax, %xmm2
	imul	%rdi, %rax
	mulsd	%xmm1, %xmm2
	mulsd	%xmm0, %xmm2
	cvtsd2si %xmm2, %rcx
	imul	%rdx, %rcx
	sub	%rcx, %rax
	mov	%rax, %rcx
	sub     %rdx, %rcx
	jl      sqr
	mov     %rcx, %rax
sqr:
	imul	%rdi, %rdi
	mulsd	%xmm1, %xmm1
	mulsd	%xmm0, %xmm1
	cvtsd2si %xmm1, %rcx
	shr	%rsi
	imul	%rdx, %rcx
	sub	%rcx, %rdi
	mov	%rdi, %rcx
	sub     %rdx, %rcx
	jl	test
	mov     %rcx, %rdi
test:
	test	%rsi, %rsi
	jnz	loop

#ifdef _WIN64
	pop	%rsi
	pop	%rdi
#endif
	ret
